/* -*- mode: C++; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

template <uint N, typename PointAlloc> 
_HD_ Point<N,PointAlloc>::Point() {}

/************************************************************************************************/

//Construtor using 3 string
template <uint N, typename PointAlloc> 
_H_ Point<N,PointAlloc>::Point(const std::string& x_str,const std::string& y_str,const std::string& z_str)  : _x(x_str), _y(y_str), _z(z_str){}


/************************************************************************************************/

//assignment operator
template <uint N, typename PointAlloc> 
template <typename PointAlloc2>
_HD_ Point<N,PointAlloc>& Point<N,PointAlloc>::operator=(const Point<N,PointAlloc2>& po) {
	//if(&po == this) 
	//	return *this;
	//else {
		_x = po._x;
		_y = po._y; 
		_z = po._z;
		return *this;
		//}
}

/************************************************************************************************/

//CLASS Elliptic Curve (ECurve)

// default constructor
template <uint N, typename AllocGFP, typename AllocA>
_HD_ ECurve<N, AllocGFP, AllocA>::ECurve(): _a(), _Fp() {}

// constructor of the curve from parameter a, b and the prime
template <uint N, typename AllocGFP, typename AllocA>
ECurve<N, AllocGFP, AllocA>::ECurve(string a, string p) : _a(a), _Fp(p) {}


/************************************************************************************************/

// addition of 2 points on the curve, compute p1 = p1 + p2
// useful when optimized square
template <uint N, typename AllocGFP, typename AllocA>
template<typename PointAlloc1, typename PointAlloc2>
_HD_ void ECurve<N, AllocGFP, AllocA>::add_in(Point<N, PointAlloc1>& p1, const Point<N, PointAlloc2>& p2) {
	_var1.init(0);
	_var2.init(0);
	_var3.init(0);
	_Fp.mul(_var1, p1._z, p1._z); //_var1 = Z1^2
	_Fp.mul(_var2, p2._z, p2._z); //_var2 = Z2^2

	_Fp.mul(_var3, p1._x, _var2); //_var3 = X1*Z2^2 = U1
	_Fp.mul(p1._y, p1._y, p2._z); //Y1 = Y1 * Z2
	_Fp.mul(p1._y, p1._y, _var2); //Y1 = Y1 * Z2 * Z2^2 = S1
	p1._x = p1._z;
	_Fp.add_in(p1._x, p2._z);     //X1 = Z1+Z2
	_Fp.mul(p1._x, p1._x, p1._x); //X1 = (Z1+Z2)^2
	_Fp.sub_in(p1._x, _var2);     //X1 = (Z1+Z2)^2 - Z2^2
	_Fp.mul(_var2, p2._x, _var1); //_var2 = X2 * Z1^2  = U2
	_Fp.sub_in(p1._x, _var1);     //X1 = (Z1+Z2)^2 - Z2^2 - Z1^2
	_Fp.mul(_var1, _var1, p1._z); //_var1 = Z1 * Z1^2
	_Fp.mul(_var1, _var1, p2._y); //_var1 = Z1 * Z1^2 * Y2 = S2

	_Fp.sub_in(_var1,p1._y);      //_var1 = (Z1 * Z1^2 * Y2) - (Z2 * Z2^2 * Y1) = S2-S1
	_Fp.mul(_var1, 2, _var1);     //_var1 = 2*((Z1 * Z1^2 * Y2) - (Z2 * Z2^2 * Y1)) = r

	_Fp.sub_in(_var2, _var3);     //_var2 = X2*Z1^2 - X1*Z2^2 = H
	//Z3
	_Fp.mul(p1._z, p1._x, _var2); //Z1 = ((Z1+Z2)^2 - Z2^2 - Z1^2) * (X2*Z1^2 - X1*Z2^2)

	_Fp.mul(p1._x, 2, _var2);     //X1 = 2 * (X2*Z1^2 - X1*Z2^2)
	_Fp.mul(p1._x, p1._x, p1._x); //X1 = (2 * (X2*Z1^2 - X1*Z2^2))^2 = I
	
	_Fp.mul(_var3, _var3, p1._x); //_var3 = (2 * (X2*Z1^2 - X1*Z2^2))^2 * (X1*Z2^2) = V
	_Fp.mul(_var2, _var2, p1._x); //_var2 = (X2*Z1^2 - X1*Z2^2) * (2 * (X2*Z1^2 - X1*Z2^2))^2  = J

	_Fp.mul(p1._x, _var1, _var1); //X1 = r^2
	_Fp.sub_in(p1._x, _var2);     //X1 = r^2 - J
	_Fp.sub_in(p1._x, _var3);     //X1 = r^2 - J - V
	//X3
	_Fp.sub_in(p1._x, _var3);     //X1 = r^2 - J - 2*V
	
	_Fp.mul(_var2, _var2, p1._y); //_var2 = J*S1
	_Fp.mul(_var2, 2, _var2);     //_var2 = 2*J*S1
	p1._y = _var3;
	_Fp.sub_in(p1._y, p1._x);     //Y1 = V - X3
	_Fp.mul(p1._y, p1._y, _var1); //Y1 = r * (V - X3)
	//Y3
	_Fp.sub_in(p1._y, _var2);     //Y1 = r * (V - X3) - 2 * J * S1
}


/************************************************************************************************/

// addition of 2 points on the curve, compute p1 = p1 + p2
//useful when no optimized square
template <uint N, typename AllocGFP, typename AllocA>
template<typename PointAlloc1, typename PointAlloc2>
_HD_ void ECurve<N, AllocGFP, AllocA>::add_in2(Point<N, PointAlloc1>& p1, const Point<N, PointAlloc2>& p2) {
	if(p1._z == 0) {
		p1 = p2;
	}
	else {
		_var1.init(0);
		_var2.init(0);
		_var3.init(0);

		_Fp.mul(_var1, p1._z, p1._z); //_var1 = Z1^2
		_Fp.mul(_var2, p2._z, p2._z); //_var2 = Z2^2

		_Fp.mul(_var3, p1._x, _var2); //_var3 = X1*Z2^2 = U1
		_Fp.mul(p1._y, p1._y, p2._z); //Y1 = Y1 * Z2
		_Fp.mul(p1._y, p1._y, _var2); //Y1 = Y1 * Z2 * Z2^2 = S1

		_Fp.mul(p1._x, p1._z, p2._z); //X1 = Z1+Z2
		_Fp.mul(_var2, p2._x, _var1); //_var2 = X2 * Z1^2  = U2
		_Fp.mul(_var1, _var1, p1._z); //_var1 = Z1 * Z1^2
		_Fp.mul(_var1, _var1, p2._y); //_var1 = Z1 * Z1^2 * Y2 = S2

		_Fp.sub_in(_var1,p1._y);      //_var1 = (Z1 * Z1^2 * Y2) - (Z2 * Z2^2 * Y1) = S2-S1
		_Fp.sub_in(_var2, _var3);     //_var2 = X2*Z1^2 - X1*Z2^2 = H
		//Z3
		_Fp.mul(p1._z, p1._x, _var2); //Z1 = ((Z1+Z2)^2 - Z2^2 - Z1^2) * (X2*Z1^2 - X1*Z2^2)
		_Fp.mul(p1._x,_var2, _var2); //X1 = (2 * (X2*Z1^2 - X1*Z2^2))^2 = I
	
		_Fp.mul(_var3, _var3, p1._x); //_var3 = (2 * (X2*Z1^2 - X1*Z2^2))^2 * (X1*Z2^2) = V
		_Fp.mul(_var2, _var2, p1._x); //_var2 = (X2*Z1^2 - X1*Z2^2) * (2 * (X2*Z1^2 - X1*Z2^2))^2  = J

		_Fp.mul(p1._x, _var1, _var1); //X1 = r^2
		_Fp.sub_in(p1._x, _var2);     //X1 = r^2 - J
		_Fp.sub_in(p1._x, _var3);     //X1 = r^2 - J - V
		//X3
		_Fp.sub_in(p1._x, _var3);     //X1 = r^2 - J - 2*V
	
		_Fp.mul(_var2, _var2, p1._y); //_var2 = J*S1
		p1._y = _var3;
		_Fp.sub_in(p1._y, p1._x); //Y1 = V - X3
		_Fp.mul(p1._y, p1._y, _var1); //Y1 = r * (V - X3)
		//Y3
		_Fp.sub_in(p1._y, _var2);     //Y1 = r * (V - X3) - 2 * J * S1
	}
}

/************************************************************************************************/

//doubling a point on the curve, compute p1 = 2*p1
// useful when optimized square
template <uint N, typename AllocGFP, typename AllocA>
template<typename PointAlloc1>
_HD_ void ECurve<N, AllocGFP, AllocA>::dbl_in(Point<N,PointAlloc1>& p) {
	_var1.init(0);
	_var2.init(0);
	_Fp.mul(_var1, p._z, p._z);   //_var1 = Z^2
	_Fp.add_in(p._z, p._y);       //Z = Z+Y
	_Fp.mul(p._z, p._z, p._z);    //Z = (Z+Y)^2
	
	_Fp.mul(p._y, p._y, p._y);    //Y = Y^2
	_Fp.sub_in(p._z, p._y);       //Z = (Z+Y)^2-Y^2
	_Fp.sub_in(p._z, _var1);      //Z = (Z+Y)^2-Y^2-Z^2

	_Fp.mul(_var1, _var1, _var1); //_var1 = Z^4
	_Fp.mul(_var1, _var1, _a);    //_var1 = a*Z^4
	_var2 = p._x;
	_Fp.add_in(_var2, p._y);      //_var2 = X+Y^2
	_Fp.mul(_var2, _var2, _var2); //_var2 = (X+Y^2)^2
	_Fp.mul(p._x, p._x, p._x);    //X = X^2
	_Fp.sub_in(_var2, p._x);      //_var2 = (X+Y^2)^2 - X^2
	_Fp.mul(p._y, p._y, p._y);    //Y = Y^4
	_Fp.sub_in(_var2, p._y);      //_var2 = (X+Y^2)^2 - X^2 - Y^4
	_Fp.mul(_var2, 2, _var2);     //_var2 = 2*((X+Y^2)^2 - X^2 - Y^4)

	_Fp.mul(p._x, 3, p._x);       //X = 3*X^2
	_Fp.add_in(p._x, _var1);      //X = 3*X^2 + a*Z^4
	
	_Fp.mul(p._y, 8, p._y);       //Y = 8*Y^4
	_Fp.mul(_var1, p._x, p._x);   //_var1 = (3*X^2 + a*Z^4)^2
	_Fp.sub_in(_var1, _var2);     //_var1 = (3*X^2 + a*Z^4)^2 - 2*((X+Y^2)^2 - X^2 - Y^4)
	_Fp.sub_in(_var1, _var2);     //_var1 = (3*X^2 + a*Z^4)^2 - 2*2*((X+Y^2)^2 - X^2 - Y^4)
	
	_Fp.sub_in(_var2, _var1);     //_var2 = 2*((X+Y^2)^2 - X^2 - Y^4) - ((3*X^2 + a*Z^4)^2 - 2*2*((X+Y^2)^2 - X^2 - Y^4))
	_Fp.mul(_var2, _var2, p._x);  //_var2 =(2*((X+Y^2)^2 - X^2 - Y^4)-((3*X^2+a*Z^4)^2-2*2*((X+Y^2)^2-X^2-Y^4)))*(3*X^2+a*Z^4)
	_Fp.sub_in(_var2, p._y);      //_var2=((2*((X+Y^2)^2-X^2-Y^4)-((3*X^2+a*Z^4)^2-4*((X+Y^2)^2-X^2-Y^4)))*(3*X^2+a*Z^4))-8*Y1^2
	
	p._y = _var2;
	p._x = _var1;
}

/************************************************************************************************/

//doubling a point on the curve, compute p1 = 2*p1
// useful when no optimized square
template <uint N, typename AllocGFP, typename AllocA>
template<typename PointAlloc1>
_HD_ void ECurve<N, AllocGFP, AllocA>::dbl_in2(Point<N,PointAlloc1>& p) {
	_var1.init(0);
	_var2.init(0);
	_Fp.mul(_var1, p._z, p._z);                 //_var1 = Z^2
	_Fp.mul(p._z, p._z, p._y);                  // z = z*y
	_Fp.mul(p._z, 2, p._z);                     // z = 2*y*z
	_Fp.mul(_var1, _var1, _var1);               //_var1 = Z^4
	_Fp.mul(_var1, _a, _var1);                  //_var1 = a*Z^4

	_Fp.mul(p._y, p._y, p._y);                  //y = y^2
	_Fp.mul(_var2,p._x, p._y);                  //var2 = x*y^2
	_Fp.mul(_var2, 4, _var2);                   //var2 = 4*x*y^2

	_Fp.mul(p._x,p._x, p._x);                   //x = x^2
	_Fp.mul(p._x, 3, p._x);                     //x = 3*x^2

	_Fp.add_in(p._x, _var1);                    //x = 3*x^2+a*z^4
	
	_Fp.mul(p._y, p._y, p._y);                  //Y = Y^2
	
	_Fp.mul(p._y, 8, p._y);                     //Y = 8*Y^4
	_Fp.mul(_var1, p._x, p._x);                 //_var1 = (3*X^2 + a*Z^4)^2
	_Fp.sub_in(_var1, _var2);                   //_var1 = (3*X^2 + a*Z^4)^2 - 2*((X+Y^2)^2 - X^2 - Y^4)
	_Fp.sub_in(_var1, _var2);                   //_var1 = (3*X^2 + a*Z^4)^2 - 2*2*((X+Y^2)^2 - X^2 - Y^4)
	
	_Fp.sub_in(_var2, _var1);                   //_var2 = 2*((X+Y^2)^2 - X^2 - Y^4) - ((3*X^2 + a*Z^4)^2 - 2*2*((X+Y^2)^2 - X^2 - Y^4))
	_Fp.mul(_var2, _var2, p._x);                //_var2 =(2*((X+Y^2)^2 - X^2 - Y^4)-((3*X^2+a*Z^4)^2-2*2*((X+Y^2)^2-X^2-Y^4)))*(3*X^2+a*Z^4)
	_Fp.sub_in(_var2, p._y);                    //_var2=((2*((X+Y^2)^2-X^2-Y^4)-((3*X^2+a*Z^4)^2-4*((X+Y^2)^2-X^2-Y^4)))*(3*X^2+a*Z^4))-8*Y1^2
	
	p._y = _var2;
	p._x = _var1;
}

/************************************************************************************************/

//compute Q = 2Q+P
template <uint N, typename AllocGFP, typename AllocA>
template<typename PointAlloc1, typename PointAlloc2>
_HD_ void ECurve<N, AllocGFP, AllocA>::dbl_add_in(Point<N, PointAlloc1>& Q, const Point<N, PointAlloc2>& P) {
	_var1.init(0);
	_var2.init(0);
	_var3.init(0);
	_Fp.mul(_var1, Q._z, Q._z);   //_var1 = Z^2
	_Fp.mul(Q._z, Q._z, Q._y);    //z = z*y
	_Fp.mul(Q._z, 2, Q._z);       //Z3 = 2*y*z
	_Fp.mul(_var1, _var1, _var1); //_var1 = Z^4
	_Fp.mul(_var1, _a, _var1);    //_var1 = a*Z^4

	_Fp.mul(Q._y, Q._y, Q._y);    //y = y^2
	_Fp.mul(_var2,Q._x, Q._y);    //var2 = x*y^2
	_Fp.mul(_var2, 4, _var2);     //var2 = 4*x*y^2

	_Fp.mul(Q._x,Q._x, Q._x);     //x = x^2
	_Fp.mul(Q._x, 3, Q._x);       //x = 3*x^2

	_Fp.add_in(Q._x, _var1);      //x = 3*x^2+a*z^4
	
	_Fp.mul(Q._y, Q._y, Q._y);    //Y = Y^2
	
	_Fp.mul(Q._y, 8, Q._y);       //Y = 8*Y^4
	_Fp.mul(_var1, Q._x, Q._x);   //_var1 = (3*X^2 + a*Z^4)^2
	_Fp.sub_in(_var1, _var2);     //_var1 = (3*X^2 + a*Z^4)^2 - 2*((X+Y^2)^2 - X^2 - Y^4)
	_Fp.sub_in(_var1, _var2);     // X = _var1 = (3*X^2 + a*Z^4)^2 - 2*2*((X+Y^2)^2 - X^2 - Y^4)
	
	_Fp.sub_in(_var2, _var1);     //_var2 = 2*((X+Y^2)^2 - X^2 - Y^4) - ((3*X^2 + a*Z^4)^2 - 2*2*((X+Y^2)^2 - X^2 - Y^4))
	_Fp.mul(_var2, _var2, Q._x);  // _var2 =(2*((X+Y^2)^2 - X^2 - Y^4)-((3*X^2+a*Z^4)^2-2*2*((X+Y^2)^2-X^2-Y^4)))*(3*X^2+a*Z^4)
	_Fp.sub_in(_var2, Q._y);      // Y3 = _var2=((2*((X+Y^2)^2-X^2-Y^4)-((3*X^2+a*Z^4)^2-4*((X+Y^2)^2-X^2-Y^4)))*(3*X^2+a*Z^4))-8*Y1^2
	
	Q._y = _var2;
	Q._x = _var1;

	_Fp.mul(_var1, Q._z, Q._z); //_var1 = Z1^2
	_Fp.mul(_var2, P._z, P._z); //_var2 = Z2^2

	_Fp.mul(_var3, Q._x, _var2); //_var3 = X1*Z2^2 = U1
	_Fp.mul(Q._y, Q._y, P._z); //Y1 = Y1 * Z2
	_Fp.mul(Q._y, Q._y, _var2); //Y1 = Y1 * Z2 * Z2^2 = S1

	_Fp.mul(Q._x, Q._z, P._z); //X1 = Z1+Z2
	_Fp.mul(_var2, P._x, _var1); //_var2 = X2 * Z1^2  = U2
	_Fp.mul(_var1, _var1, Q._z); //_var1 = Z1 * Z1^2
	_Fp.mul(_var1, _var1, P._y); //_var1 = Z1 * Z1^2 * Y2 = S2

	_Fp.sub_in(_var1,Q._y);      //_var1 = (Z1 * Z1^2 * Y2) - (Z2 * Z2^2 * Y1) = S2-S1
	_Fp.sub_in(_var2, _var3);     //_var2 = X2*Z1^2 - X1*Z2^2 = H
	//Z3
	_Fp.mul(Q._z, Q._x, _var2); //Z1 = ((Z1+Z2)^2 - Z2^2 - Z1^2) * (X2*Z1^2 - X1*Z2^2)
	_Fp.mul(Q._x,_var2, _var2); //X1 = (2 * (X2*Z1^2 - X1*Z2^2))^2 = I
	
	_Fp.mul(_var3, _var3, Q._x); //_var3 = (2 * (X2*Z1^2 - X1*Z2^2))^2 * (X1*Z2^2) = V
	_Fp.mul(_var2, _var2, Q._x); //_var2 = (X2*Z1^2 - X1*Z2^2) * (2 * (X2*Z1^2 - X1*Z2^2))^2  = J

	_Fp.mul(Q._x, _var1, _var1); //X1 = r^2
	_Fp.sub_in(Q._x, _var2);     //X1 = r^2 - J
	_Fp.sub_in(Q._x, _var3);     //X1 = r^2 - J - V
	//X3
	_Fp.sub_in(Q._x, _var3);     //X1 = r^2 - J - 2*V
	
	_Fp.mul(_var2, _var2, Q._y); //_var2 = J*S1
	Q._y = _var3;
	_Fp.sub_in(Q._y, Q._x); //Y1 = V - X3
	_Fp.mul(Q._y, Q._y, _var1); //Y1 = r * (V - X3)
	//Y3
	_Fp.sub_in(Q._y, _var2);     //Y1 = r * (V - X3) - 2 * J * S1

}

/************************************************************************************************/ 
// scalar multiplication [k]P over the curve
template <uint N, typename AllocGFP, typename AllocA>
template<typename Alloc, typename PointAlloc1, typename PointAlloc2>
_D_ void ECurve<N, AllocGFP, AllocA>::scalarMulLeftToRight(Point<N, PointAlloc1>& p_res, const Integer<N,Alloc>& k, const Point<N,PointAlloc2>& P) {
	p_res = P;
	//compte number of bits
	int nb_bits = 16*(NB_LIMB-1)+Utils::log2ceil(k[NB_LIMB-1]);	
	for(int i = nb_bits-2; i >= 0; i--) {
		//doubling
		dbl_in2(p_res);
		if(k.get_bit_at(i) == 1) {
			//if k[i] == 1, addition
			add_in2(p_res, P);
		}
	}
}

/************************************************************************************************/

// scalar multiplication [k]P over the curve
template <uint N, typename AllocGFP, typename AllocA>
template<typename Alloc, typename PointAlloc1, typename PointAlloc2>
_D_ void ECurve<N, AllocGFP, AllocA>::scalarMulRightToLeft(Point<N, PointAlloc1>& p_res, const Integer<N,Alloc>& k, Point<N,PointAlloc2>& P) {
	int nb_bits = 16*(NB_LIMB-1)+Utils::log2ceil(k[NB_LIMB-1]);	
	p_res._z = 0;

#pragma unroll 
	for(int j = 0; j < N; j++) {
		if(k.get_bit_at(j) == 1) {
		//if k[i] == 1, addition
			add_in2(p_res, P);
		}
		//doubling
		if(j < nb_bits) dbl_in2(P);
	}
}

/************************************************************************************************/

#ifdef USE_NAF
template <uint N, typename AllocGFP, typename AllocA>
template<typename PointAlloc1, typename PointAlloc2>
_HD_ void ECurve<N, AllocGFP, AllocA>::sub_in2(Point<N, PointAlloc1>& p1, const Point<N, PointAlloc2>& p2) {
	if(p1._z == 0) {
		p1 = p2;
	}
	else {
		_var1.init(0);
		_var2.init(0);
		_var3.init(0);
		_var4.init(0);

		_Fp.sub_in(_var4, p2._y); // NAF: -y2

		_Fp.mul(_var1, p1._z, p1._z); //_var1 = Z1^2
		_Fp.mul(_var2, p2._z, p2._z); //_var2 = Z2^2

		_Fp.mul(_var3, p1._x, _var2); //_var3 = X1*Z2^2 = U1
		_Fp.mul(p1._y, p1._y, p2._z); //Y1 = Y1 * Z2
		_Fp.mul(p1._y, p1._y, _var2); //Y1 = Y1 * Z2 * Z2^2 = S1

		_Fp.mul(p1._x, p1._z, p2._z); //X1 = Z1+Z2
		_Fp.mul(_var2, p2._x, _var1); //_var2 = X2 * Z1^2  = U2
		_Fp.mul(_var1, _var1, p1._z); //_var1 = Z1 * Z1^2

		//		_Fp.mul(_var1, _var1, p2._y); //_var1 = Z1 * Z1^2 * Y2 = S2
		_Fp.mul(_var1, _var1, _var4); // NAF

		_Fp.sub_in(_var1,p1._y);      //_var1 = (Z1 * Z1^2 * Y2) - (Z2 * Z2^2 * Y1) = S2-S1
		_Fp.sub_in(_var2, _var3);     //_var2 = X2*Z1^2 - X1*Z2^2 = H
		//Z3
		_Fp.mul(p1._z, p1._x, _var2); //Z1 = ((Z1+Z2)^2 - Z2^2 - Z1^2) * (X2*Z1^2 - X1*Z2^2)
		_Fp.mul(p1._x,_var2, _var2); //X1 = (2 * (X2*Z1^2 - X1*Z2^2))^2 = I
	
		_Fp.mul(_var3, _var3, p1._x); //_var3 = (2 * (X2*Z1^2 - X1*Z2^2))^2 * (X1*Z2^2) = V
		_Fp.mul(_var2, _var2, p1._x); //_var2 = (X2*Z1^2 - X1*Z2^2) * (2 * (X2*Z1^2 - X1*Z2^2))^2  = J

		_Fp.mul(p1._x, _var1, _var1); //X1 = r^2
		_Fp.sub_in(p1._x, _var2);     //X1 = r^2 - J
		_Fp.sub_in(p1._x, _var3);     //X1 = r^2 - J - V
		//X3
		_Fp.sub_in(p1._x, _var3);     //X1 = r^2 - J - 2*V
	
		_Fp.mul(_var2, _var2, p1._y); //_var2 = J*S1
		p1._y = _var3;
		_Fp.sub_in(p1._y, p1._x); //Y1 = V - X3
		_Fp.mul(p1._y, p1._y, _var1); //Y1 = r * (V - X3)
		//Y3
		_Fp.sub_in(p1._y, _var2);     //Y1 = r * (V - X3) - 2 * J * S1
	}
}

template <uint N, typename AllocGFP, typename AllocA>
template<typename PointAlloc1>
_HD_ void ECurve<N, AllocGFP, AllocA>::smul_in(Point<N, PointAlloc1>& p, int d) 
{
	if (d<0)
		{
			for (uint i=0; i<-d+1; i++)
				sub_in2(p, p);
		}
	else
		{
			for (uint i=0; i<d-1; i++)
			        add_in2(p, p);
		}
}

template <uint N, typename AllocGFP, typename AllocA>
template<typename Alloc, typename PointAlloc1, typename PointAlloc2>
_D_ void ECurve<N, AllocGFP, AllocA>::scalarMulLeftToRightNaf(Point<N, PointAlloc1>& p_res, const Integer<N,Alloc>& k, const Point<N,PointAlloc2>& P) {
       	short d;

	_naf.from_integer(k);

	d = _naf.get_digit_at(_naf._msd_rank);

#ifdef NAF3
	_P3 = P;
	for (int i=0; i<2; i++)
		add_in2(_P3, P);
	if (d > 0)
		{
			if (d == 1)
				p_res = P;
			else // d == 3
				p_res = _P3;
		}
	else // d < 0  (d !=0 for msd)
		{
			p_res = P;
			if (d == -1)
				for (int i=0; i<2; i++)
					sub_in2(p_res, P);
			else // d == -3
				for (int i=0; i<4; i++)
					sub_in2(p_res, P);				
		}
#else // NAF2
	p_res = P;
	if (d==-1)
		for (int i=0; i<2; i++)
			sub_in2(p_res, P);
#endif

#pragma unroll 
	for (int i = _naf._msd_rank-1; i >= 0; i--) 
		{
			dbl_in2(p_res);
			
			d = _naf.get_digit_at(i);
#ifdef NAF3
			if (d != 0)
				{
					if (d > 0)
						{
							if (d == 1)
								add_in2(p_res, P);
							else // d == 3
								add_in2(p_res, _P3);
						}
					else // d < 0
						{
							if (d == -1)
								sub_in2(p_res, P);
							else // d == -3
								sub_in2(p_res, _P3);
						}
				}
#else
			if (d != 0)
				{
					if (d == 1)
						add_in2(p_res, P);
					else // d == -1
						sub_in2(p_res, P);
				}
#endif
		}

}

#endif
