/* -*- mode: C++; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

#ifndef __IntegerReg__H
#define __IntegerReg__H

#include <cuda.h>
#include <iostream>
#include <sstream>
#include <vector>
using namespace std;

#include "gpu_memory.cuh"
#include "integer.cuh"

#define BASE 32
#define MASK 0xFFFFFFFF

typedef unsigned long uint64;

template<uint N>
void stringToArray(uint *T, string s){
	//number of digit in base 10 in one limb
	int pow_base = ceil(log(pow(2.,BASE))/log(10))-1;
	//maximal number in base 10 which can be store in one limb
	uint64 bbase    = pow(10.,pow_base);

	//size of string
	uint64 n_size   = s.size();
	//firsts characters of the string (in case where the string size is not a multiple of pow_base)
	uint64 first_c  = n_size % pow_base;
	uint64 _used = 0;
	//get the first_c characters and convert into int
	if(first_c) {
		string limb = s.substr(0,first_c);
		uint l = 0;
		sscanf(limb.c_str(),"%d",&l);
		T[0]=l;
		_used  = 1;
	}
	int index = first_c;
	//while the string isn't totally parsed, compute integer*10000+part of string (corresponding to 10*bbase+part of string of size 4)
	while(index != n_size) {
		//get the substring of size "pow_base" and convert into integer
		string limb = s.substr(index,pow_base);
		uint64 l = 0;
		sscanf(limb.c_str(),"%d",&l);
		
		if(_used) {
			uint64 carry = 0;
			//compute Integer*bbase (10000)
			for(int i = 0; i < _used; i++) {
				uint64 mul = T[i]*bbase;
				mul+=carry;
				T[i]= mul & MASK;
				carry    = mul >> BASE;
			}
			if(carry != 0) {
				T[_used]=carry;
				_used++;
			}
			carry = l;
			int i = 0;
			//compute Integer+part of string
			while(carry) {				
				uint64 add = T[i]+carry;
				T[i]=add & MASK;
				carry    = add >> BASE;
				i++;
			}
			if(carry != 0) {
				T[_used]= carry;
				_used++;
			}
		}		
		else {
			T[0]=l;
			_used = 1;
		}
		index += pow_base;
	}
	for(int i = _used; i < N; i++) 
		T[i]=0;
}


template<uint N>
string arrayToString(uint *T)  {
	//number of digit (base 10) which can be represented from one mp_d
	int p10 = floor(log(1UL << BASE)/log(10.));
	//number of limb in base 10 needed to represent the Integer in base 10
	int nb_limb = ceil(BASE*N*log(2)/log(10));
	nb_limb = (nb_limb % p10 ? nb_limb/p10+1 : nb_limb/p10);
	uint64 tab[nb_limb];
	//	for(int i = 0; i < nb_limb; i++) tab[i] = 0;
	uint index = 0;
	uint64 pow10 = pow(10.,p10);
	//each step compute (Number in base 10)*2^16+new mp_d
	for(int i = N-1; i >= 0; i--) {
		//compute Number in base 10 * 2^16
		uint64 carry = 0;
		for(int j = 0; j < index; j++) {
			uint64 mul = (tab[j] << BASE)+carry;
			tab[j] = mul % pow10;
			carry = mul / pow10;
		}        
		while(carry != 0) {
			tab[index] = carry % pow10;
			carry = carry / pow10;
			index++;
		}
		carry = T[i];
		int k = 0;
		//compute Number in base 10 + following limb
		while (k < index && carry != 0) {
			uint64 add = tab[k]+carry;
			tab[k] = add % pow10;
			carry = add / pow10;
			k++;
		}
		while(carry) {
			tab[index] = carry % pow10;
			carry = carry / pow10;
			index++;
		}		
	}

	ostringstream os;
	if (index == 0)
		return string("0");
	else {
		os << tab[index-1];
		for(int i = index-2; i >= 0; i--) {
			ostringstream t;
			t << tab[i];
			int s = t.str().size();
			while(s != p10) {
				os << 0;
				s++;
			}			
			os << t.str();
		}
		return os.str();
	}
}



/****************************************************************************************************/
/*****************               INTEGER STRUCTURE                             **********************/
/****************************************************************************************************/

template <uint N>
struct IntegerReg;

template<>
struct IntegerReg<1> {
	uint limb;
	
	// defaut constructor (initialize to 0)
	__device__ IntegerReg()  {limb=0;}
  
	// create integer from array
	__device__ IntegerReg(uint *T){limb=*T;}

	// copy integer to array
	__device__ void set(uint* T) { *T=limb; }

	// set to constant one
	__device__ void set_one(){limb=1;}

	// set to constant zero
	__device__ void set_zero(){limb=0;}
};
template <uint N>
struct IntegerReg : IntegerReg<N-1> {
	uint limb;
  
	// defaut constructor (initialize to 0)
	__device__ IntegerReg() : IntegerReg<N-1>() {limb=0;}

	 // create integer from array
	__device__ IntegerReg(uint *T) : IntegerReg<N-1>(T) {limb=T[N-1];}

	 // copy integer to array
	__device__ void set(uint* T) {T[N-1]=limb; IntegerReg<N-1>::set(T); }

	// set to constant one
	__device__ void set_one(){limb=0;IntegerReg<N-1>::set_one();}

	// set to constant one
	__device__ void set_zero(){limb=0;IntegerReg<N-1>::set_zero();}

 };

 // limb accessor
#define AT(__v__,__i__)  __v__. IntegerReg<(__i__)+1>::limb

// bit accessor
#define BIT_AT(__v__,__i__) (AT(__v__,(__i__)/16)>>((__i__%16)))&0X1

template<uint N>
ostream& operator<<(ostream& os, IntegerReg<N>& x)
{
  uint T[N];
  x.set(T);
  return os<<arrayToString<N>(T);
}



/****************************************************************************************************/
/*****************                   MEMORY CLASS                              **********************/
/****************************************************************************************************/
#ifndef PAD
#define PAD(N,pad)  (N+(N % pad ? N-(N % pad) : 0))
#endif
 template<uint beg, uint end, uint NBTHREADS>
 struct gMemoryLoop {

	 template<uint N>
	 __device__ static void load(IntegerReg<N>& x, uint* v, uint k)  
	 {
		 AT(x,beg) = v[beg*PAD(NBTHREADS,16)+k];		 
		 gMemoryLoop<beg+1,end,NBTHREADS>::load(x,v,k);
	 }
	 template<uint N>
	 __device__ static void store(uint*v, const IntegerReg<N>& x, uint k)  
	 {
		 v[beg*PAD(NBTHREADS,16)+k] = AT(x,beg);			
		 gMemoryLoop<beg+1,end,NBTHREADS>::store(v,x,k);
	 }

 };

 template<uint end, uint NBTHREADS>
 struct gMemoryLoop<end,end,NBTHREADS> {

	 template<uint N>
	 __device__ static void load(IntegerReg<N>& x, uint* v, uint k)  
	 {
		 AT(x,end) = v[end*PAD(NBTHREADS,16)+k];					
	 }
	 template<uint N>
	 __device__ static void store(uint*v, const IntegerReg<N>& x, uint k)  
	 {
		 v[end*PAD(NBTHREADS,16)+k] = AT(x,end);			
	 }
 };

 template<uint N, uint size>
 class MMemory {
 public:
	 __host__ static void sendToGPU(uint*& v, const vector<uint*>& vec)
	 {
		 const uint pad = PAD(size,16);
		 uint linear_padded_data[N*pad];
		 for(uint i = 0; i < size; i++) 
			 for(int j = 0; j < N; j++) 
				 linear_padded_data[j*pad+i] = vec[i][j];

		 cudaMalloc((void**) &v, sizeof(uint)*pad*N);
		 cudaMemcpy(v,linear_padded_data, sizeof(uint)*pad*N, cudaMemcpyHostToDevice);
	 }

	 __host__ static void sendToGPU(uint*& v, const vector<uint>& vec)
	 {
		 const uint pad = PAD(size,16);
		 uint linear_padded_data[pad];
		 for(uint i = 0; i < size; i++) 
			 linear_padded_data[i] = vec[i];
		 cudaMalloc((void**) &v, sizeof(uint)*pad);
		 cudaMemcpy(v,linear_padded_data, sizeof(uint)*pad, cudaMemcpyHostToDevice);
	 } 

	 __host__ static void sendToCPU(vector<uint*>& vec, uint* v)
	 {
		 const uint pad = PAD(size,16);		
		 uint mp[sizeof(uint)*pad*N];		
		 cudaMemcpy(mp,v,sizeof(uint)*pad*N, cudaMemcpyDeviceToHost);		
		 for(uint i = 0; i < size; i++) 
			 for(uint j = 0; j < N; j++)
				 (vec[i])[j] = mp[j*pad+i];			
	 }

	 __device__ static void globalLoad(IntegerReg<N>& x, uint* v, uint k)
	 {
		 gMemoryLoop<0,N-1,size>::load(x,v,k);
	 }

	 __device__ static void globalStore( uint* v,const IntegerReg<N>& x, uint k)
	 {
		 gMemoryLoop<0,N-1,size>::store(v,x,k);
	 }
 };


/****************************************************************************************************/
/*****************             INTEGER OPERATIONS                              **********************/
/****************************************************************************************************/

 template<uint beg, uint end>
 struct IntegerLoop {

	 template<uint N>
	 __device__ static bool lessThan(const IntegerReg<N>& a, const IntegerReg<N>& b) 
	 { return AT(a,end)<AT(b,end) || (AT(a,end)==AT(b,end) && IntegerLoop<beg,end-1>::lessThan(a,b)); }

	 template<uint N>
	 __device__ static bool moreThan(const IntegerReg<N>& a, const IntegerReg<N>& b)
	 { return AT(a,end)>AT(b,end) || (AT(a,end)==AT(b,end) && IntegerLoop<beg,end-1>::moreThan(a,b)); }

	 template<uint N>
	 __device__ static bool egal(const IntegerReg<N>& a, const IntegerReg<N>& b)
	 { return  AT(a,beg) == AT(b,beg) && IntegerLoop<beg+1,end>::egal(a,b); }

	 template<uint N>
	 __device__ static bool egalInt(const IntegerReg<N>& a, int b)
	 { return  AT(a,beg) == b &&  IntegerLoop<beg+1,end>::egalInt(a, 0); }

	 template<uint N>
	 __device__ static bool lessThanInt (const IntegerReg<N>& a, int b)
	 { return  IntegerLoop<1,end>::egalInt(a,0) && (AT(a,0) < b);} 

	 template<uint N>
	 __device__ static bool moreThanInt (const IntegerReg<N>& a, int b)
	 { return  IntegerLoop<1,end>::egalInt(a,0) && (AT(a,0) > b); }


 };

 template<uint end>
 struct IntegerLoop<end,end> {

	 template<uint N>  __device__ static bool lessThan(const IntegerReg<N>& a, const IntegerReg<N>& b) { return AT(a,end)<AT(b,end);}

	 template<uint N>  __device__ static bool moreThan(const IntegerReg<N>& a, const IntegerReg<N>& b) { return AT(a,end)>AT(b,end);}

	 template<uint N>  __device__ static bool egal (const IntegerReg<N>& a, const IntegerReg<N>& b) { return AT(a,end)==AT(b,end); }

	 template<uint N>  __device__ static bool egalInt (const IntegerReg<N>& a, int b) { return AT(a,end)== b; }

 };

 // integer vs integer comparison
 template<uint N>
 __device__ bool operator<= (const IntegerReg<N>& a, const IntegerReg<N>& b){ return !IntegerLoop<0,N-1>::moreThan(a,b);}
 template<uint N>
 __device__ bool operator>= (const IntegerReg<N>& a, const IntegerReg<N>& b){ return !IntegerLoop<0,N-1>::lessThan(a,b);}
 template<uint N>
 __device__ bool operator== (const IntegerReg<N>& a, const IntegerReg<N>& b){ return IntegerLoop<0,N-1>::egal(a,b);}
 template<uint N>
 __device__ bool operator<  (const IntegerReg<N>& a, const IntegerReg<N>& b){ return IntegerLoop<0,N-1>::lessThan(a,b);}
 template<uint N>
 __device__ bool operator>  (const IntegerReg<N>& a, const IntegerReg<N>& b){ return IntegerLoop<0,N-1>::moreThan(a,b);}
 template<uint N>
 __device__ bool operator!= (const IntegerReg<N>& a, const IntegerReg<N>& b){ return !IntegerLoop<0,N-1>::egal(a,b);}

 // integer vs int comparison
 template<uint N>
 __device__ bool operator<= (const IntegerReg<N>& a, int b){return !IntegerLoop<0,N-1>::moreThanInt(a,b);}
 template<uint N>
 __device__ bool operator>= (const IntegerReg<N>& a, int b){return !IntegerLoop<0,N-1>::lessThanInt(a,b);}
 template<uint N>
 __device__ bool operator== (const IntegerReg<N>& a, int b){return IntegerLoop<0,N-1>::egalInt(a,b);}
 template<uint N>
 __device__ bool operator<  (const IntegerReg<N>& a, int b){return IntegerLoop<0,N-1>::lessThanInt(a,b);}
 template<uint N>
 __device__ bool operator>  (const IntegerReg<N>& a, int b){return IntegerLoop<0,N-1>::moreThanInt(a,b);}
 template<uint N>
 __device__ bool operator!= (const IntegerReg<N>& a, int b){return !IntegerLoop<0,N-1>::egalInt(a,b);}


/********************
  * UTIITY FUNCTION *
  *******************/

// a = a + b + ci -> return carry if a > 2^BASE  
 __device__ inline uint addin_c(uint& a, const uint& b, uint ci=0)
{
	a+=(b+ci);
	return a<(b+ci);
}

// a = b + d + ci -> return carry if a > 2^BASE  
 __device__ inline uint add_c(uint& a, const uint& b, const uint& d, uint ci=0)
{
	a=b+d+ci;
	return a<(b+ci);
}


// a*b = l_ab + 2^BASE * h_ab
 __device__ inline void fullmul(const uint& a, const uint& b, uint& l_ab, uint& h_ab)
{
	uint tmp,c=0;
	l_ab  = __umul24(a&0xFFFF, b&0xFFFF);
	h_ab  = __umul24(a>>16,b>>16 );
	tmp   = __umul24(a>>16,b&0xFFFF);
	c     = addin_c(l_ab, (tmp&0xFFFF)<<16,c);
	c     = addin_c(h_ab, tmp>>16,c);	
	tmp   = __umul24(a&0xFFFF, b>>16);
	c     = addin_c(l_ab,(tmp&0xFFFF)<<16,c);
	c     = addin_c (h_ab,tmp>>16,c);	
}

// return floor( (al+2^32*ah) / p )
 __device__ inline void div(const uint& al, const uint& ah, const uint& p, uint &q)
{
	float t=4294967296.;
	t=t*ah+al;
	q= t/ (float) p;	
}


 /*********************
  * SUBTRACTION MOD P *
  *********************/

 // Main loop
 template<uint beg, uint end>
 struct subLoop {
	 template<uint N>
	  __device__ static void subin_g(IntegerReg<N>& a, const IntegerReg<N>& b, const IntegerReg<N>& prime, uint carry=0)
	 {		 
		 uint sub= AT(a,beg)-AT(b,beg)-carry;
		 carry= AT(a,beg) < sub;
		 AT(a,beg)=sub;
		 subLoop<beg+1,end>::subin_g(a,b,prime,carry);
	 }  

	 template<uint N>
	 __device__  static void subin_l(IntegerReg<N>& a, const IntegerReg<N>& b, const IntegerReg<N>& prime, uint carry=0)
	 {
		 AT(a,beg)+=AT(prime,beg)+carry;
		 carry= AT(a,beg) < (AT(prime,beg)+carry);		 
		 uint sub=AT(a,beg) - AT(b,beg);
		 carry-= AT(a,beg) < sub;
		 AT(a,beg)=sub;		 
		 subLoop<beg+1,end>::subin_l(a,b,prime,carry);
	 }  
 };

 // End of loop
 template<uint end>
 struct subLoop<end,end> {
	 template<uint N>
	 __device__  static void subin_g(IntegerReg<N>& a, const IntegerReg<N>& b, const IntegerReg<N>& prime, uint carry=0)
	 {
		 uint sub= AT(a,end)-AT(b,end)-carry;
		 carry= AT(a,end) < sub;
		 AT(a,end)=sub;			 
	 }  

	 template<uint N>
	 __device__  static void subin_l(IntegerReg<N>& a, const IntegerReg<N>& b, const IntegerReg<N>& prime, uint carry=0)
	 {
		 AT(a,end)+=AT(prime,end)+carry;
		 carry= AT(a,end) < (AT(prime,end)+carry);		 
		 uint sub= AT(a,end) - AT(b,end);
		 carry-= AT(a,end) < sub;
		 AT(a,end)=sub;		 
	 }    
 };
 /*****************/

/*******************
  * ADDITION MOD P *
  ******************/

 // Main loop
 template<uint beg, uint end>
 struct addLoop {
	 template<uint N>
	  __device__ static void addin(IntegerReg<N>& a, const IntegerReg<N>& b, const IntegerReg<N>& prime, uint &carry)
	 {
		 AT(a,beg)+=AT(b,beg)+carry;		 		 
		 carry=  AT(a,beg) < (AT(b,beg)+carry);
		 addLoop<beg+1,end>::addin(a,b,prime,carry);		 
	 }	
 };

 // End of loop
 template<uint end>
 struct addLoop<end,end> {
	 template<uint N>
	  __device__ static void addin(IntegerReg<N>& a, const IntegerReg<N>& b, const IntegerReg<N>& prime, uint &carry)
	 {
		 AT(a,end)+=AT(b,end)+carry;		 		 
		 carry=  AT(a,end) < (AT(b,end)+carry);
		 if (a >= prime || carry)
			 subLoop<0,end>::subin_g(a,prime,prime);		 		 		 
	 }
 };

 /*****************/


 /************************
  * MULTIPLICATION MOD P *
  ************************/

 // Main loop
 template<uint beg, uint end>
 struct mulLoop {
	 template<uint N>
	 __device__ static void mul(IntegerReg<N>& res, const IntegerReg<N>& op1, const IntegerReg<N>& op2, const IntegerReg<N>& prime, IntegerReg<N>&var1, const uint& minvp, uint carryA=0)
	 {
		 uint xiyl, xiyh;
		 fullmul(AT(op1,beg),AT(op2,0),xiyl,xiyh); 

		 uint ui; // ui = (a0+xi*yo)*invp mod 2^32;
		 ui=AT(var1,0)+xiyl;    
		 ui*=minvp;
		 
		 uint uiml, uimh;
		 fullmul(ui,AT(prime,0), uiml,uimh);

		 uint carryl; // carryl = (var0 + xi*y0 + ui* prime0) / 2^32
		 uint carryh; // carryh = (xi*y0)/ 2^32 + (ui* prime0)/ 2^32  + carryl
		 carryh =add_c  (carryl, xiyl, uiml);
		 carryh+=addin_c(carryl,AT(var1,0));				 
		 carryl = addin_c(carryh,xiyh);
		 carryl+= addin_c(carryh,uimh);
		 
		 mulLoop<1,end>::mul_inner(op2,prime,var1,ui,AT(op1,beg),carryl,carryh);
		 
		 AT(var1,N-1) = carryh + carryA; 
		 mulLoop<beg+1,end>::mul(res,op1,op2,prime,var1,minvp,carryl);   
	 }  
	 template<uint N>
	 __device__ static void mul_inner(const IntegerReg<N>& op2, const IntegerReg<N>& prime,  IntegerReg<N>& var1, uint& ui, const uint& xi, uint& carryl, uint &carryh )
	 {		 
		 
		 carryh = carryl+add_c(AT(var1,beg-1), carryh, AT(var1, beg)) ;
		 uint xiyil, xiyih;
		 fullmul(xi,AT(op2,beg), xiyil,xiyih);		
		 uint uiml, uimh;
		 fullmul(ui, AT(prime,beg), uiml,uimh);
		 
		 carryh += addin_c(AT(var1,beg-1),xiyil);

		 carryh += addin_c(AT(var1,beg-1),uiml);

		 carryl  = addin_c(carryh, xiyih);
		 carryl += addin_c(carryh,uimh);
		    
		 mulLoop<beg+1,end>::mul_inner(op2,prime,var1,ui,xi,carryl, carryh);
	 } 
	 template<uint N>
	 __device__ static void mulInt(IntegerReg<N>& res, int& op1, const IntegerReg<N>& op2, const IntegerReg<N>& prime, IntegerReg<N>&var1, uint carryh=0)
	 {
		 uint mulh,mull;
		 fullmul(abs(op1), AT(op2,beg), mull,mulh); 		 
		 carryh = add_c (AT(var1,beg),mull,carryh);
		 carryh += mulh; // carry is possible only if op1 is 2^32-1 		 
		 mulLoop<beg+1,end>::mulInt(res,op1,op2,prime,var1,carryh);
	 }	
	 template<uint N>
	 __device__ static void mulInt_2(IntegerReg<N>& var1, const IntegerReg<N>& prime, uint q, uint carryh=0)
	 {
		 uint mull,mulh;
		 fullmul(q,AT(prime,beg),mull,mulh);		 
		 uint sub= AT(var1,beg)- mull -carryh;
		 carryh  = AT(var1,beg) < sub;
		 carryh += mulh;
		 AT(var1,beg)= sub;
		 mulLoop<beg+1,end>::mulInt_2(var1, prime, q, carryh);		
	 }
 };
 // End of loop
 template<uint end>
 struct mulLoop<end,end> {
	 template<uint N>
	  __device__ static void mul(IntegerReg<N>& res, const IntegerReg<N>& op1, const IntegerReg<N>& op2, const IntegerReg<N>& prime, IntegerReg<N>&var1, const uint& minvp,uint carryA=0)
	 {
		 uint xiyl, xiyh;		 
		 fullmul(AT(op1,end), AT(op2,0),xiyl,xiyh); 
		 		 
		 uint ui; // ui = (a0+xi*yo)*invp mod 2^32;
		 ui=AT(var1,0)+xiyl;    
		 ui*=minvp;
			 
		 uint uiml, uimh;
		 fullmul(ui,AT(prime,0), uiml,uimh);

		 uint carryl; // carryl = var0 + xi*y0 + ui* prime0 mod 2^32
		 uint carryh; // carryh = (xi*y0)/ 2^32 + (ui* prime0)/ 2^32  + carryl		
		 carryh =add_c  (carryl, xiyl, uiml);
		 carryh+=addin_c(carryl,AT(var1,0));		
		 carryl = addin_c(carryh,xiyh);
		 carryl+= addin_c(carryh,uimh);		 
	
		 mulLoop<1,end>::mul_inner(op2,prime,var1,ui,AT(op1,end),carryl,carryh);
		 AT(var1,N-1) = carryh + carryA;
		 
		 if (carryl || var1 >= prime)
			 subLoop<0,N-1>::subin_g(var1,prime,prime); // use subin_g to not rewrite the code (possible because not mod p in this case)
		 
		 res=var1;
	 }    
	 template<uint N>
	 __device__ static void mul_inner(const IntegerReg<N>& op2, const IntegerReg<N>& prime,  IntegerReg<N>& var1, uint &ui, const uint &xi, uint& carryl, uint &carryh )
	 {				 
		 carryh = carryl+add_c(AT(var1,end-1), carryh, AT(var1, end)) ;
		 
		 uint xiyil, xiyih;
		 fullmul(xi,AT(op2,end), xiyil,xiyih);				 		 
		 uint uiml, uimh;
		 fullmul(ui, AT(prime,end), uiml,uimh);

		 carryh += addin_c(AT(var1,end-1),xiyil);		
		 carryh += addin_c(AT(var1,end-1),uiml);				
		 carryl  = addin_c(carryh, xiyih);
		 carryl += addin_c(carryh,uimh);
	 }
	 template<uint N>
	 __device__ static void mulInt(IntegerReg<N>& res, int& op1, const IntegerReg<N>& op2, const IntegerReg<N>& prime, IntegerReg<N>& var1, uint carryh=0)
	 {
		 uint mulh,mull;
		 fullmul(abs(op1), AT(op2,end), mull,mulh);		 
		 carryh = add_c (AT(var1,end),mull,carryh);
		 carryh +=mulh;
		 uint q; div(AT(var1,N-1),carryh,AT(prime,N-1),q);
		 
		 mulLoop<0,end>::mulInt_2(var1, prime, q);		
		 if (op1 < 0){
			 res=prime;
			 subLoop<0,N-1>::subin_g(res,var1,prime);
		 }
		 else
			 res=var1;
		 

	 }	
	 template<uint N>
	 __device__  static void mulInt_2(IntegerReg<N>& var1, const IntegerReg<N>& prime, uint q, uint carryh=0)
	 {
		 uint mull,mulh;
		 fullmul(q,AT(prime,end),mull,mulh);		 
		 uint sub= AT(var1,end)- mull -carryh;
		 carryh  = AT(var1,end) < sub;
		 carryh += mulh;
		 AT(var1,end)= sub;
	 }		
 };
 /*****************/


 // FUNCTION: addition mod p
 template<uint N> __device__ void addin(IntegerReg<N>& a, const IntegerReg<N>& b, const IntegerReg<N>& prime){ uint carry=0; addLoop<0,N-1>::addin(a,b,prime,carry);}


 // FUNCTION: subtraction mod p
 template<uint N>  __device__ void subin(IntegerReg<N>& a, const IntegerReg<N>& b, const IntegerReg<N>& prime)
 {
	 if (a>=b) subLoop<0,N-1>::subin_g(a,b,prime);else subLoop<0,N-1>::subin_l(a,b,prime);
 }

 // FUNCTION: multiplication mod p
 template<uint N>  __device__  void mul(IntegerReg<N>& a, const IntegerReg<N>& b, const IntegerReg<N>& c, const IntegerReg<N>& prime, uint minvp, IntegerReg<N> &var)
 {
	 var.set_zero();
	 mulLoop<0,N-1>::mul(a,b,c,prime,var,minvp);
 }

 // FUNCTION: multiplication mod p by an int
template<uint N>  __device__ void mul(IntegerReg<N>& a, int b, const IntegerReg<N>& c, const IntegerReg<N>& prime, IntegerReg<N> &var)
 {
	 mulLoop<0,N-1>::mulInt(a,b,c,prime,var);
 }



/*****************************
 *** Montgomery conversion ***
 *****************************/

template<uint N>
__device__ void convertToMontg(IntegerReg<N>& aR, const IntegerReg<N>& a, const IntegerReg<N>& prime, IntegerReg<N> &var)
{	
	mul(aR, 65536, a, prime,var);
	for (uint i=0;i<2*N-1;i++)
		mul(aR, 65536, aR, prime,var);	
}

template<uint N>
__device__ void convertFromMontg(IntegerReg<N>& a, const IntegerReg<N>& aR, const IntegerReg<N>& prime, uint invp, const IntegerReg<N> &one, IntegerReg<N> &var2)
{
	mul(a,one,aR,prime,invp,var2);	
}


/****************************************************************************************************/
/*****************      OPERATION ON POINT OVER A ELL CURVE                    **********************/
/****************************************************************************************************/

 template<uint N>
 struct PointReg { 
	 IntegerReg<N> _x;
	 IntegerReg<N> _y;
	 IntegerReg<N> _z;
 };
/*
 template<uint N> 
 __device__ void dblin(PointReg<N>& p, const IntegerReg<N>& _a, const IntegerReg<N>& prime, uint minvp, IntegerReg<N>& _var1, IntegerReg<N>& _var2, IntegerReg<N>& _var3)
 {
	 mul(_var1, p._z, p._z,prime, minvp,_var3);        

	 mul(p._z, p._z, p._y,prime,minvp,_var3);   //
	 mul(p._z, 2, p._z,prime,_var3);  
  	 
	 mul(_var1, _var1, _var1,prime,minvp,_var3);//     
	 mul(_var1, _a, _var1,prime,minvp,_var3);   //
	 mul(p._y, p._y, p._y,prime,minvp,_var3);   //
      
	 mul(_var2,p._x, p._y,prime,minvp,_var3);         
	 mul(_var2, 4, _var2,prime,_var3);          
	 mul(p._x,p._x, p._x,prime,minvp,_var3);    //       
	 mul(p._x, 3, p._x,prime,_var3);            
	 addin(p._x, _var1,prime);           
	 mul(p._y, p._y, p._y,prime,minvp,_var3);   //        
	 mul(p._y, 8, p._y,prime,_var3);           
	 mul(_var1, p._x, p._x,prime,minvp,_var3);  
	 subin(_var1, _var2,prime);          
	 subin(_var1, _var2,prime);           
	 subin(_var2, _var1,prime); 
	 mul(_var2, _var2, p._x,prime,minvp,_var3); //
	 subin(_var2, p._y,prime);           
	 p._y = _var2;
	 p._x = _var1;
 }
*/
 template<uint N> 
 __device__ void dblin(PointReg<N>& p, const IntegerReg<N>& _a, const IntegerReg<N>& prime, uint minvp, IntegerReg<N>& _var1, IntegerReg<N>& _var2, IntegerReg<N>& _var3)
 {
	 mul(_var1, p._z, p._z,prime, minvp,_var3);        

	 mul(_var2, p._z, p._y,prime,minvp,_var3);  
	 mul(p._z, 2, _var2 ,prime,_var3);  
  	 
	 mul(_var2, _var1, _var1,prime,minvp,_var3);
	 mul(_var1, _a, _var2,prime,minvp,_var3);   
	 mul(p._y, p._y, p._y,prime,minvp,_var3);   /**/
      
	 // var2 appear
	 mul(_var2,p._x, p._y,prime,minvp,_var3);         
	 mul(_var2, 4, _var2,prime,_var3);   
       
	 mul(p._x, p._x, p._x,prime,minvp,_var3);    /**/       

	 mul(p._x, 3, p._x,prime,_var3);            
	 addin(p._x, _var1,prime);     
      
	 mul(p._y, p._y, p._y,prime,minvp,_var3);   /**/ 
       
	 mul(p._y, 8, p._y,prime,_var3);           

	 mul(_var1, p._x, p._x,prime,minvp,_var3);  
	 subin(_var1, _var2,prime);          
	 subin(_var1, _var2,prime);           
	 subin(_var2, _var1,prime); 
	 p._x = _var1;
	 mul(_var1, _var2, p._x,prime,minvp,_var3); 
	 subin(_var1, p._y,prime);           
	 p._y = _var1;
	 
 }
/*
template<uint N> 
__device__ void addin(PointReg<N>& p1, const PointReg<N>& p2,  const IntegerReg<N>& prime, uint minvp, 
				   IntegerReg<N>& _var1, IntegerReg<N>& _var2, IntegerReg<N>& _var3, IntegerReg<N>& _var4)
 {	
	if(p1._z == 0) {
		p1 = p2;
	}
	else {
		mul(_var1, p1._z, p1._z,prime,minvp,_var4); 
		mul(_var2, p2._z, p2._z,prime,minvp,_var4); 
		mul(_var3, p1._x, _var2,prime,minvp,_var4); 
		mul(p1._y, p1._y, p2._z,prime,minvp,_var4); 
		mul(p1._y, p1._y, _var2,prime,minvp,_var4); 
		mul(p1._x, p1._z, p2._z,prime,minvp,_var4); 
		mul(_var2, p2._x, _var1,prime,minvp,_var4); 
		mul(_var1, _var1, p1._z,prime,minvp,_var4); 
		mul(_var1, _var1, p2._y,prime,minvp,_var4); 
		subin(_var1,p1._y,prime);      
		subin(_var2, _var3,prime);     		
		mul(p1._z, p1._x, _var2,prime,minvp,_var4); 
		mul(p1._x, _var2, _var2,prime,minvp,_var4); 	
		mul(_var3, _var3, p1._x,prime,minvp,_var4); 
		mul(_var2, _var2, p1._x,prime,minvp,_var4); 
		mul(p1._x, _var1, _var1,prime,minvp,_var4); 
		subin(p1._x, _var2,prime);     
		subin(p1._x, _var3,prime);     		
		subin(p1._x, _var3,prime);     	
		mul(_var2, _var2, p1._y,prime,minvp,_var4); 
		p1._y = _var3;
		subin(p1._y, p1._x,prime);  
		mul(p1._y, p1._y, _var1,prime,minvp,_var4); 		
		subin(p1._y, _var2,prime);     
	}
 }
*/
template<uint N> 
__device__ void addin(PointReg<N>& p1, const PointReg<N>& p2,  const IntegerReg<N>& prime, uint minvp, 
				   IntegerReg<N>& _var1, IntegerReg<N>& _var2, IntegerReg<N>& _var3, IntegerReg<N>& _var4)
 {	
	if(p1._z == 0) {
		p1 = p2;
	}
	else {
		mul(_var1, p1._z, p1._z,prime,minvp,_var4); 
		mul(_var2, p2._z, p2._z,prime,minvp,_var4); 
		mul(_var3, p1._x, _var2,prime,minvp,_var4); 

		mul(p1._y, p1._y, p2._z,prime,minvp,_var4); //
		mul(p1._y, p1._y, _var2,prime,minvp,_var4); //

		mul(p1._x, p1._z, p2._z,prime,minvp,_var4); 
		mul(_var2, p2._x, _var1,prime,minvp,_var4); 
		mul(_var1, _var1, p1._z,prime,minvp,_var4); //
		mul(_var1, _var1, p2._y,prime,minvp,_var4); //
		subin(_var1,p1._y,prime);      
		subin(_var2, _var3,prime);     		
		mul(p1._z, p1._x, _var2,prime,minvp,_var4); 
		mul(p1._x, _var2, _var2,prime,minvp,_var4); 	
		mul(_var3, _var3, p1._x,prime,minvp,_var4); //
		mul(_var2, _var2, p1._x,prime,minvp,_var4); //
		mul(p1._x, _var1, _var1,prime,minvp,_var4); 
		subin(p1._x, _var2,prime);     
		subin(p1._x, _var3,prime);     		
		subin(p1._x, _var3,prime);     	
		mul(_var2, _var2, p1._y,prime,minvp,_var4); //
		p1._y = _var3;
		subin(p1._y, p1._x,prime);  
		mul(p1._y, p1._y, _var1,prime,minvp,_var4); //		
		subin(p1._y, _var2,prime);     
	}
 }


template<uint beg, uint end>
struct kpLoop {
	template<uint N>
	__device__ static void mul(PointReg<N>& Q, const IntegerReg<N>& k, PointReg<N>& P, const IntegerReg<N>& _a, const IntegerReg<N>& prime,  uint minvp,
				   IntegerReg<N>& _var1, IntegerReg<N>& _var2, IntegerReg<N>& _var3, IntegerReg<N>& _var4 )
	{
		if (BIT_AT(k,beg)==1)
			addin(Q,P,prime,minvp, _var1,_var2,_var3,_var4);		
		dblin(P,_a,prime,minvp, _var1,_var2,_var3);
		kpLoop<beg+1,end>::mul(Q,k,P,_a,prime,minvp, _var1,_var2,_var3,_var4);
		
	}
};

template<uint end>
struct kpLoop<end,end> {
	template<uint N>
	__device__ static void mul(PointReg<N>& Q, const IntegerReg<N>& k, PointReg<N>& P, const IntegerReg<N>& _a, const IntegerReg<N>& prime,  uint minvp, 
				   IntegerReg<N>& _var1, IntegerReg<N>& _var2, IntegerReg<N>& _var3, IntegerReg<N>& _var4 )
	{
		if (BIT_AT(k,end)==1)
			addin(Q,P,prime,minvp, _var1,_var2,_var3,_var4);
	}
};

template<uint N> 
__device__ void scalarMul(PointReg<N>& Q, const IntegerReg<N>& k, PointReg<N>& P,  const IntegerReg<N>& _a, const IntegerReg<N>& prime, uint minvp)
{
	IntegerReg<N> var1,var2,var3,var4;
	kpLoop<0,32*N-1>::mul(Q,k,P,_a,prime,minvp,var1,var2,var3,var4);

}
 
template<uint N> 
__device__ void scalarMul(PointReg<N>& Q, const Integer<32*N>& k, PointReg<N>& P,  IntegerReg<N>& _a, const IntegerReg<N>& prime, uint minvp,IntegerReg<N>& var1,IntegerReg<N>& var2,IntegerReg<N>& var3,IntegerReg<N>& var4)
{
	//IntegerReg<N> var1,var2,var3,var4;

	// convert a, Px, Py, Pz in montgomery form
	//convertToMontg(_a  ,_a  , prime, var1);
	//convertToMontg(P._x,P._x, prime, var1);
	//convertToMontg(P._y,P._y, prime, var1);
	//convertToMontg(P._z,P._z, prime, var1);
	
	for(ushort j = 0; j < 32*N; j++) {
		if(k.get_bit_at(j) == 1) 
			addin(Q,P,prime,minvp, var1,var2,var3,var4);						
		dblin(P,_a,prime,minvp,var1,var2,var3);
	}	
	//var1.set_one();
	//convertFromMontg(Q._x,Q._x, prime, minvp, var1, var2);
	//convertFromMontg(Q._y,Q._y, prime, minvp, var1, var2);
	//convertFromMontg(Q._z,Q._z, prime, minvp, var1, var2);	
}



#endif
