/* -*- mode: C++; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

#ifndef INTEGER_GPU_CUH_
#define  INTEGER_GPU_CUH_
#include <string>
#include <sstream>
#include <iostream>
using namespace std;
#include <cuda.h>

#include "utils.hpp"

// define the number of LIMB
#define NB_LIMB (N & 0x10 ? (N >> 4)+1 : (N >> 4))

/*** define struct for allocation plan ***/

// define the texture on GPU
texture<ushort,2> GPUTexture; 


// plan for texture memory (idx set the static position of the integer in the texture GPUTexture define globally in gpu_memory.cuh file)
template <uint idx> struct TextureMemory{};

// plan for local memory
struct LocalMemory{};


/*** define the allocator ***/
template<uint N, typename memory>
struct Allocator;
 

/*** specialization of the allocator for local memory ***/
template<uint N>
struct Allocator<N, LocalMemory> {

	ushort _mp_d[NB_LIMB];

	//_HD_ uint& get(uint i) {return _mp_d[i];}
	_HD_ ushort& get(uint i) {return _mp_d[i];}

	//_HD_ uint  get(uint i) const {return _mp_d[i];}
	_HD_ ushort  get(uint i) const {return _mp_d[i];}
 	
	//_HD_ void  set(uint i, const uint& x) {_mp_d[i]=x;}
	_HD_ void  set(uint i, const ushort& x) {_mp_d[i]=x;}
};

/*** specialization of the allocator for texture memory ***/
uint dumb_memory=17;

template<uint N, uint idx>
struct Allocator<N, TextureMemory<idx> > {

	_HD_ uint& get(uint i) {return dumb_memory;/*impossible texture integer must be always constant  */}

	// !!! considered that texture memory is duplicater for each thread (each thread stores 5 integers in texture)
 	_D_ uint  get(uint i) const {return tex2D(GPUTexture,i,Utils::tid()*5+idx);} 
	
 	_HD_ void  set(uint i, const uint& x) {/* impossible texture is read only */ }
};



template<uint N, typename Alloc=LocalMemory >
class Integer {
public:
	Allocator<N,Alloc> _alloc;
			
	_HD_ Integer(){}

	_H_ Integer(string s){
		//number of digit in base 10 in one limb
		int pow_base = ceil(log(pow(2.,16))/log(10))-1;
		//maximal number in base 10 which can be store in one limb
		int bbase    = pow(10.,pow_base);

		//size of string
		int n_size   = s.size();
		//firsts characters of the string (in case where the string size is not a multiple of pow_base)
		int first_c  = n_size % pow_base;
		int _used = 0;
		//get the first_c characters and convert into int
		if(first_c) {
			string limb = s.substr(0,first_c);
			int l = 0;
			sscanf(limb.c_str(),"%d",&l);
			_alloc.set(0,l);
			_used  = 1;
		}
		int index = first_c;
		//while the string isn't totally parsed, compute integer*10000+part of string (corresponding to 10*bbase+part of string of size 4)
		while(index != n_size) {
			//get the substring of size "pow_base" and convert into integer
			string limb = s.substr(index,pow_base);
			int l = 0;
			sscanf(limb.c_str(),"%d",&l);
		
			if(_used) {
				uint carry = 0;
				//compute Integer*bbase (10000)
				for(int i = 0; i < _used; i++) {
					uint mul = _alloc.get(i)*bbase+carry;
					_alloc.set(i, mul & 0xFFFF);
					carry    = mul >> 16;
				}
				if(carry != 0) {
					_alloc.set(_used, carry);
					_used++;
				}
				carry = l;
				int i = 0;
				//compute Integer+part of string
				while(carry) {				
					uint add = _alloc.get(i)+carry;
					_alloc.set(i, add & 0xFFFF);
					carry    = add >> 16;
					i++;
				}
				if(carry != 0) {
					_alloc.set(_used, carry);
					_used++;
				}
			}		
			else {
				_alloc.set(0, l);
				_used = 1;
			}
			index += pow_base;
		}
		for(int i = _used; i < NB_LIMB; i++) _alloc.set(i,0);
	}


	_H_ string to_string() const {
		//number of digit (base 10) which can be represented from one mp_d
		int p10 = floor(log(1 << 16)/log(10.));
		//number of limb in base 10 needed to represent the Integer in base 10
		int nb_limb = ceil(N*log(2)/log(10));
		nb_limb = (nb_limb % p10 ? nb_limb/p10+1 : nb_limb/p10);
		uint tab[nb_limb];
		//	for(int i = 0; i < nb_limb; i++) tab[i] = 0;
		uint index = 0;
		int pow10 = pow(10.,p10);
		//each step compute (Number in base 10)*2^16+new mp_d
		for(int i = NB_LIMB-1; i >= 0; i--) {
			//compute Number in base 10 * 2^16
			uint carry = 0;
			for(int j = 0; j < index; j++) {
				uint mul = (tab[j] << 16)+carry;
				tab[j] = mul % pow10;
				carry = mul / pow10;
			}
	

			while(carry != 0) {
				tab[index] = carry % pow10;
				carry = carry / pow10;
				index++;
			}
			carry = _alloc.get(i);
			int k = 0;
			//compute Number in base 10 + following limb
			while (k < index && carry != 0) {
				uint add = tab[k]+carry;
				tab[k] = add % pow10;
				carry = add / pow10;
				k++;
			}
			while(carry) {
				tab[index] = carry % pow10;
				carry = carry / pow10;
				index++;
			}
		}
		//
		ostringstream os;
		if (index == 0)
			return string("0");
		else {
			os << tab[index-1];
			for(int i = index-2; i >= 0; i--) {
				ostringstream t;
				t << tab[i];
				int s = t.str().size();
				while(s != p10) {
					os << 0;
					s++;
				}
				os << t.str();
			}
			return os.str();
		}
	}


	_D_ void init(int n=0)
	{ 
		_alloc.set(0, n & ((1 << 16)-1));
		_alloc.set(1, n >> 16);	
		for(int i = (n >> 16 ? 2 : 1); i < NB_LIMB; i++) _alloc.set(i,0);
	}
	
	_D_ int get_bit_at(int pos) const { return (_alloc.get(pos >> 4) >> (pos % 16) & 1);}

	_H_ uint get_hamming_weight() const {
		uint hW = 0;
		uint c;
		uint v;
		for(uint i =0; i < NB_LIMB; i++) {
			v = _alloc.get(i);
			for(c = 0; v; c++) v &= v - 1;
			hW += c;
		}
		return hW;
	}
	
	template<typename Allocator2>
	_D_ bool operator<  (const Integer<N, Allocator2>& in) const {return !(*this >= in);}

	template<typename Allocator2>
	_D_ bool operator>  (const Integer<N, Allocator2>& in) const {return !(*this <= in);}

	template<typename Allocator2>
	_D_ bool operator<= (const Integer<N, Allocator2>& in) const
	{
		int i = NB_LIMB-1;
		while (i >= 0 &&_alloc.get(i) == in._alloc.get(i)) i--;
		return (i == -1) || _alloc.get(i) < in._alloc.get(i);
	}
	
	template<typename Allocator2>
	_D_ bool operator>= (const Integer<N, Allocator2>& in) const
	{
		int i = NB_LIMB-1;
		while (i >= 0 && _alloc.get(i) == in._alloc.get(i)) i--;
		return (i == -1) || _alloc.get(i) > in._alloc.get(i);
	}
	
	template<typename Allocator2>
	_D_ bool operator== (const Integer<N, Allocator2>& in) const
	{
		int i = NB_LIMB-1;		
		while (i >= 0 && _alloc.get(i) == in._alloc.get(i)) i--;
		return (i == -1);
	}
	
	template<typename Allocator2>
	_D_ bool operator!= (const Integer<N, Allocator2>& in) const {return !(*this == in);}

	
	_D_ bool operator<  (int in) const {return !(*this >= in);}
	_D_ bool operator>  (int in) const {return !(*this <= in);}

	_D_ bool operator<= (int in) const {
		int i = NB_LIMB-1;
		while(i >= 0 && _alloc.get(i) == 0) i--;
		return (i < 2 && ((_alloc.get(1) < in >> 16) || (_alloc.get(1) == in >> 16) && (_alloc.get(0)) <= in & 0xFFFF ));
	}				

	_D_ bool operator>= (int in) const {
		int i = NB_LIMB-1;
		while(i >= 0 && _alloc.get(i) == 0) i--;
		return (i >= 2 || i < 2 && ((_alloc.get(1) > in >> 16) || _alloc.get(1) == in >> 16 && (_alloc.get(0) >= in & 0xFFFF)));
	}

	_D_ bool operator== (int in) const {
		int i = NB_LIMB-1;
		while(i >= 0 && _alloc.get(i) == 0) i--;
		return (i < 2 && _alloc.get(0) == (0xFFFF & in) && (in >> 16) == _alloc.get(1));
	}

	_D_ bool operator!= (int in) const {return !(*this == in);}

	template<typename Allocator2>
	_D_ Integer& operator=(const Integer<N,Allocator2>& op)
	{
		//if(this != &op){
#pragma unroll
			for(int i = 0; i < NB_LIMB; i++) {
				_alloc.set(i, op[i]);
			}
			//}
		return *this;
	}
	
	_D_ Integer& operator= (int n) {
		_alloc.set(0,  n & ((1 << 16)-1));
		_alloc.set(1,  n >> 16);
		for(int i = (n >> 16 ? 2 : 1); i < NB_LIMB; i++) _alloc.set(i,0);
		return *this;
	}

	//	_HD_ uint& operator[] (int i) { return _alloc.get(i);}
	_HD_ ushort& operator[] (int i) { return _alloc.get(i);}

	//	_HD_ uint operator[]  (int i) const { return _alloc.get(i);}
	_HD_ ushort operator[]  (int i) const { return _alloc.get(i);}

	//	_D_ uint get(int i) const {return _alloc.get(i);}
	_D_ uint get(int i) const {return _alloc.get(i);}

	_D_ bool is_odd() const { return (_alloc.get(0)&1)==1; }  

	// mods 2^n for small n (n<16)
	_D_ uint mods(uint n) const
	{
		uint r = _alloc.get(0)&((1<<n)-1);
		return ((r>n/2) ? (r-n) : r);
	}

	// dedicated test (FIXME until bug correction)
	_D_ bool is_gt0() const 
	{
		int a = 0;
#pragma unroll
 		for(int i = 0; i < NB_LIMB; i++) 
			a |= _alloc.get(i)&0xFFFF;
 		return (a!=0 ? true : false);
	}
	
	// inplace subtraction for small integer (n < 2^16)
	// WARNING this-n must be >= 0
	_D_ void sub_in(int n)
	{
		int carry = -n;
#pragma unroll
 		for(int i = 0; i < NB_LIMB; i++) 
			{
				int r = _alloc.get(i)+carry;
				if (r < 0)
					{
						carry = -1;
						r += 0x10000;
					}
				else 
					{
						if (r < 0x10000)
							{
								carry = 0;
							}
						else
							{
								carry = 1;
								r -= 0x10000;
							}
					}
				_alloc.set(i, r);
			}
 	}
	
	// inplace division by 2
	_D_ void div2()
	{
		int r0 = _alloc.get(0);
#pragma unroll
 		for(int i = 1; i < NB_LIMB; i++) 
			{
				int r1 = _alloc.get(i);
				_alloc.set(i-1, (r0>>1)|((r1&1)<<15));
				r0 = r1;
			}
		_alloc.set(NB_LIMB-1, (r0>>1));
 	}
		
};

#endif
