/* -*- mode: C++; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

#ifndef POINT_MULTIPLICATION_CUH_
#define POINT_MULTIPLICATION_CUH_

#include "gpu_memory.cuh"
#include "integer.cuh"
#include "ecc_gpu.cuh"
#include "wnaf.cuh"



/***** TEXTURE MEMORY VERSION *******************************************************************************************/

// !!!!!!
// DOES NOT WORK AS IT: NEED TO MODIFY CLASS ECURVE AND GFP TO SET prime and a parameter constant in the class
// !!!!!!

template <uint N>
struct SharedData2 {
	ECurve<N,TextureMemory<0>, TextureMemory<1> >     _ell;
	Integer<N>      _k; 
	Point<N,PointTextureMemory<2> >  _P;
	Point<N>        _Q;
};

// GPU Kernel function for multi scalar multiplication over ONE CURVE AND ONE POINT
// All paramter are stored in texture memory except (invp and k)
template <uint N>
__global__ void point_multiplication2(GPU_vector_data<N> data) {

	__shared__ uint shared_integer[BLOCK_SIZE*(sizeof(SharedData2<N>) >> 2)];

	SharedData2<N>* ptr;
	ptr= (SharedData2<N>*) &shared_integer[Utils::tid_block()*(sizeof(SharedData2<N>) >> 2)];		       
	ptr->_ell._Fp._invp= data.__v_invp[Utils::tid()];
			
	// load the integer k from global memory 
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_k, data.__v_k, Utils::tid());

	//sclar multiplication
	ptr->_ell.scalarMulLeftToRight(ptr->_Q, ptr->_k, ptr->_P);

	//convert PMon = (xMon, yMon, zMon) in Fp form (x*R*R^-1 mod prime)
	ptr->_ell._Fp.convertFromMontg(ptr->_Q._x, ptr->_Q._x);
	ptr->_ell._Fp.convertFromMontg(ptr->_Q._y, ptr->_Q._y);
	ptr->_ell._Fp.convertFromMontg(ptr->_Q._z, ptr->_Q._z);

	// store the Point Q  in global memory
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Qx, ptr->_Q._x, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Qy, ptr->_Q._y, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Qz, ptr->_Q._z, Utils::tid());		
}

// -------------------------------------------------------------------
// NAF shared memory version

template <uint N>
struct SharedDataNaf {
	ECurve<N>     _ell;
	Integer<N>      _k;
	Point<N>        _P;
	Point<N>        _Q;
};

template<uint N>
_D_ void setSharedDataNaf(SharedDataNaf<N>*& ptr,  uint* data) {
	ptr= (SharedDataNaf<N>*) &data[Utils::tid_block()*(sizeof(SharedDataNaf<N>) >> 2)];
}

template <uint N>
__global__ void point_multiplication_naf(GPU_vector_data<N> data) {

	__shared__ uint shared_integer[BLOCK_SIZE*(sizeof(SharedDataNaf<N>) >> 2)];

	SharedDataNaf<N>* ptr;
	setSharedDataNaf(ptr, shared_integer);

	// load the point P from global memory 
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_P._x, data.__v_Px, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_P._y, data.__v_Py, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_P._z, data.__v_Pz, Utils::tid());
	
	// load the parameter a of the curve and the prime of Fp from global memory 
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_ell._a,  data.__v_a, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_ell._Fp._prime, data.__v_prime, Utils::tid());
	ptr->_ell._Fp._invp= data.__v_invp[Utils::tid()];
		
		
	// load the integer k from global memory 
	Memory<N,NUMBER_THREADS>::globalLoad(ptr->_k, data.__v_k, Utils::tid());

	//Point coordinate in Montgomery form (x*R mod prime ...)
	ptr->_ell._Fp.convertToMontg(ptr->_P._x, ptr->_P._x);
	ptr->_ell._Fp.convertToMontg(ptr->_P._y, ptr->_P._y);
	ptr->_ell._Fp.convertToMontg(ptr->_P._z, ptr->_P._z);
	ptr->_ell._Fp.convertToMontg(ptr->_ell._a, ptr->_ell._a);

	//sclar multiplication
	ptr->_ell.scalarMulLeftToRightNaf(ptr->_Q, ptr->_k, ptr->_P);

	//convert PMon = (xMon, yMon, zMon) in Fp form (x*R*R^-1 mod prime)

 	ptr->_ell._Fp.convertFromMontg(ptr->_Q._x, ptr->_Q._x);
 	ptr->_ell._Fp.convertFromMontg(ptr->_Q._y, ptr->_Q._y);
 	ptr->_ell._Fp.convertFromMontg(ptr->_Q._z, ptr->_Q._z);

	// store the Point Q  in global memory
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Qx, ptr->_Q._x, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Qy, ptr->_Q._y, Utils::tid());
	Memory<N,NUMBER_THREADS>::globalStore(data.__v_Qz, ptr->_Q._z, Utils::tid());

}

#endif //POINT_MULTIPLICATION_CUH_
