/* -*- mode: C++; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

#include <cuda.h>

#include <iostream>
#include <sstream>
#include <vector>
using namespace std;

#include "gpu_memory.cuh"
#include "integer.cuh"
#include "integer-reg.cuh"
#include "utils.hpp"

/****************************************************************************************************/
/*****************        KERNEL FUNCTION ON GPU                               **********************/
/****************************************************************************************************/
 

 __device__ inline int tid() {return threadIdx.x+blockDim.x*threadIdx.y+threadIdx.z*blockDim.x*blockDim.y+blockIdx.x*blockDim.x*blockDim.y*blockDim.z+gridDim.x*blockDim.x*blockDim.y*blockDim.z*blockIdx.y;}

 template<uint N, uint NUMTHREADS>
 __global__ void kernel(ushort2* Ta, uint* Tx, uint* Ty, uint* Tz, uint* Tp, ushort2* Tk, uint* minvp){

	 IntegerReg<N> a; 
	 IntegerReg<N> prime;	
	 //IntegerReg<N> k; 
	 Integer<N*32> k;
	 PointReg<N> P,Q;

	 // a for Curve defintion
	 MMemory<N,NUMTHREADS>::globalLoad(a,(uint*)Ta,tid());	
	 // prime for Curve definition
	 MMemory<N,NUMTHREADS>::globalLoad(prime,Tp,tid());
	 // k for the scalr mult
	 Memory<N*32,NUMTHREADS>::globalLoad(k,Tk,tid());

	 // point P on the Curve
	 MMemory<N,NUMTHREADS>::globalLoad(P._x,Tx,tid());
	 MMemory<N,NUMTHREADS>::globalLoad(P._y,Ty,tid());
	 MMemory<N,NUMTHREADS>::globalLoad(P._z,Tz,tid());


	 IntegerReg<N> var1,var2,var3,var4;
	 Q=P;
	 convertToMontg(a  ,a  , prime, var1);
	 convertToMontg(Q._x,Q._x, prime, var1);
	 convertToMontg(Q._y,Q._y, prime, var1);
	 convertToMontg(Q._z,Q._z, prime, var1);

	 
	 dblin(Q,a,prime,minvp[tid()],var1,var2,var3);
	 //addin(Q,P,a,prime,minvp[tid()],var1,var2,var3,var4);
	 
	 //scalarMul(Q,k,P,a,prime,minvp[tid()],var1,var2,var3,var4);
	 
	 var3.set_one();	 
	 convertFromMontg(Q._x,Q._x, prime, minvp[tid()], var3, var2);	 
	 convertFromMontg(Q._y,Q._y, prime, minvp[tid()], var3, var2);	 
	 convertFromMontg(Q._z,Q._z, prime, minvp[tid()], var3, var2);	 
	 		 	 	 	
	 MMemory<N,NUMTHREADS>::globalStore(Tx,Q._x,tid());  
	 MMemory<N,NUMTHREADS>::globalStore(Ty,Q._y,tid());  
	 MMemory<N,NUMTHREADS>::globalStore(Tz,Q._z,tid());  
}



int main(){
	
	const uint NUMTHREADS=1024;
	/*
	const uint N=8;
	string prime_str  = "114804434916157216670268400907125902210939926123376911736891363408990795293183";
	string aR_str = "35818727692110112246791219884337495308805502403405326499582077407610210068615";
	string xR_str= "11621810404147645572992006334957427616308718412382077366295476033029868053677";
	string yR_str= "95745984883152444421279714097632615676426113333802962083583322613535555384499";
	string zR_str= "987654321158978753302584101562005642330058542263652302566220598922334346753";
	string k= "832713775790282162800138390289514703686440";		
	uint minvp=303653377;
	*/
	const uint N=5;
	string prime_str = "1460136393894413189719200461297206954055752748533";
	uint   minvp = 3844049315; 
	string a_str = "1379207425801995317087219990431025425553710104595";
	string x_str = "1053516606774785403302103971916551344226513513466";
	string y_str = "222804946308694396176659250680687754508869898366";
	string z_str = "1";
	
	string aR_str = "1071373631873578347692876124174497750968865986352";
	string xR_str= "932844094751263497418665088593497842767198957827";
	string yR_str= "1210682151637636416683423010561961153135538772829";
	string zR_str= "1365243436489728484484371419076065600179794443";
	string k= "468116454650349703862014075555268983662929600739";
	
	
	const uint NBITS=32*N;
	
	uint TA[N];stringToArray<N>(TA,aR_str);  
	uint TX[N];stringToArray<N>(TX,xR_str); 
	uint TY[N];stringToArray<N>(TY,yR_str);    
	uint TZ[N];stringToArray<N>(TZ,zR_str);    
	uint TP[N];stringToArray<N>(TP,prime_str); 

	//uint TK[N];stringToArray<N>(TP,k); 
	Integer<NBITS> TK(k);
		
	vector<uint*> v_TA(NUMTHREADS,TA);
	vector<uint*> v_TX(NUMTHREADS,TX);
	vector<uint*> v_TY(NUMTHREADS,TY);
	vector<uint*> v_TZ(NUMTHREADS,TZ);
	vector<uint*> v_TP(NUMTHREADS,TP);
	//vector<uint*> v_TK(NUMTHREADS,TK);
	vector<Integer<NBITS> > v_TK(NUMTHREADS,TK);
	vector<uint>  v_MINVP(NUMTHREADS,minvp);
	
			
	uint *A,*X,*Y,*Z,*P,*MINVP;
	ushort2* K;
	MMemory<N,NUMTHREADS>::sendToGPU(A,v_TA);
	MMemory<N,NUMTHREADS>::sendToGPU(X,v_TX);
	MMemory<N,NUMTHREADS>::sendToGPU(Y,v_TY);
	MMemory<N,NUMTHREADS>::sendToGPU(Z,v_TZ);
	MMemory<N,NUMTHREADS>::sendToGPU(P,v_TP);
	Memory<NBITS,NUMTHREADS>::sendToGPU(K,v_TK);
	MMemory<N,NUMTHREADS>::sendToGPU(MINVP,v_MINVP);
	
	// TIMER START
	cudaEvent_t start, stop;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);
	cudaEventRecord(start, 0); 
	///////

	kernel<N,NUMTHREADS><<<64,16>>>((ushort2*)A,X,Y,Z,P,K,MINVP);

	// TIMER STOP
	cudaEventRecord(stop, 0);
	cudaEventSynchronize(stop);
	float elapsedTime;
	cudaEventElapsedTime(&elapsedTime, start, stop);
	cudaEventDestroy(start);
	cudaEventDestroy(stop);
	////////

	MMemory<N,NUMTHREADS>::sendToCPU(v_TX,X);
	MMemory<N,NUMTHREADS>::sendToCPU(v_TY,Y);
	MMemory<N,NUMTHREADS>::sendToCPU(v_TZ,Z);

	
	cout << "X_result = " << arrayToString<N>(v_TX[0]) << endl;
	cout << "Y_result = " << arrayToString<N>(v_TY[0]) << endl;
	cout << "Z_result = " << arrayToString<N>(v_TZ[0])<< endl;

	cout <<"Total time: "<< elapsedTime<<" ms"<< endl;
	cout <<"Time per mul: "<< elapsedTime/(NUMTHREADS) <<" ms"<< endl;
	
	return 0;
}
